/*
 * Copyright 2017 by Kappich Systemberatung Aachen
 *
 * This file is part of de.kappich.pat.testumg.
 *
 * de.kappich.pat.testumg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.kappich.pat.testumg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.kappich.pat.testumg.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.kappich.pat.testumg.util;

import com.google.common.collect.ImmutableList;
import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.dav.dav.main.TerminateConnection;
import de.bsvrz.sys.funclib.kappich.annotations.NotNull;
import de.bsvrz.sys.funclib.kappich.annotations.Nullable;

import java.nio.file.Path;
import java.util.*;
import java.util.function.Function;

/**
 *
 *
 * @author Kappich Systemberatung
 */
public class DafApplication<T> {

	private final DafApplicationEnvironment _dafApplicationEnvironment;

	private final Function<String[], T> _creator;

	private final List<String> _args;

	private final String _user;

	private final String _id;

	private ClientDavInterface _managementConnection;

	private T _instance;

	protected DafApplication(final DafApplicationEnvironment dafApplicationEnvironment, final Function<String[], T> creator, final List<String> args, final String debugLevel, final String user) {
		_dafApplicationEnvironment = dafApplicationEnvironment;
		_creator = creator;
		_args = new ArrayList<>(args);
		_user = user;
		_id = UUID.randomUUID().toString();
	}

	public T start() {
		_managementConnection = _dafApplicationEnvironment.connect();
		_instance = _creator.apply(getArguments());
		return _instance;
	}

	public ClientApplication getClientApplication() {
		try {
			List<SystemObject> applications = _managementConnection.getLocalApplicationObject().getType().getObjects();
			DataModel dataModel = _managementConnection.getDataModel();
			AttributeGroup atg = dataModel.getAttributeGroup("atg.applikationsFertigmeldung");
			Aspect asp = dataModel.getAspect("asp.standard");
			DataDescription dataDescription = new DataDescription(atg, asp);
			while(true) {
				ResultData[] data = _managementConnection.getData(applications.toArray(new SystemObject[0]), dataDescription, 10000);
				for(ResultData datum : data) {
					if(datum.hasData()) {
						String id = datum.getData().getTextValue("Inkarnationsname").getText();
						if(id.equals(_id)) {
							return (ClientApplication) datum.getObject();
						}
					}
				}
				Thread.sleep(100);
			}
		}
		catch(InterruptedException e) {
			throw new AssertionError(e);
		}
	}

	public ClientApplication waitUntilReady() {
		try {
			DataModel dataModel = _managementConnection.getDataModel();
			AttributeGroup atg = dataModel.getAttributeGroup("atg.applikationsFertigmeldung");
			Aspect asp = dataModel.getAspect("asp.standard");
			DataDescription dataDescription = new DataDescription(atg, asp);
			ClientApplication clientApplication = getClientApplication();
			while(true) {
				ResultData datum = _managementConnection.getData(clientApplication, dataDescription, 10000);
				if(datum.hasData()) {
					String ready = datum.getData().getTextValue("InitialisierungFertig").getText();
					if(ready.equals("Ja")) {
						return (ClientApplication) datum.getObject();
					}
				}
				Thread.sleep(100);
			}
		}
		catch(InterruptedException e) {
			throw new AssertionError(e);
		}
	}

	public void startAndWait() {
		start();
		getClientApplication();
	}

	public void startAndWaitUntilReady() {
		start();
		waitUntilReady();
	}

	public T getInstance() {
		return _instance;
	}

	public final void terminate() {
		sendTerminationSignal();
		_instance = null;
	}

	protected void sendTerminationSignal() {
		try {
			TerminateConnection.sendTerminationData(_managementConnection, Collections.singletonList(getClientApplication()));
		}
		catch(InterruptedException e) {
			throw new AssertionError(e);
		}
	}

	public final Path getRootDir() {
		return _dafApplicationEnvironment.getRootDir();
	}

	public void addArguments(String... arguments) {
		_args.addAll(Arrays.asList(arguments));
	}

	private String[] getArguments() {
		ImmutableList.Builder<String> builder = ImmutableList.builder();
		addConnectionArgs(builder);
		addCustomArgs(builder);
		builder.addAll(_args);
		return builder.build().toArray(new String[0]);
	}

	protected void addCustomArgs(final ImmutableList.Builder<String> builder) {

	}

	private void addConnectionArgs(final ImmutableList.Builder<String> builder) {
		builder.add("-benutzer=" + _user);
		builder.add("-inkarnationsName=" + _id);
		builder.add("-authentifizierung=" + getPasswordFilePath());
		builder.add("-datenverteiler=localhost:" + _dafApplicationEnvironment.getAppPort());
	}

	@NotNull
	private Path getPasswordFilePath() {
		return _dafApplicationEnvironment.getWorkingDirectory().resolve("passwd.properties");
	}

	public FakeParamApp getFakeParamApp() {
		return _dafApplicationEnvironment.getFakeParamApp();
	}

	public ClientDavInterface getManagementConnection() {
		return _managementConnection;
	}

	public List<String> getParameters() {
		return Arrays.asList(getArguments());
	}
}
