/*
 * Copyright 2013 by Kappich Systemberatung Aachen
 *
 * This file is part of de.kappich.pat.testumg.
 *
 * de.kappich.pat.testumg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.kappich.pat.testumg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.kappich.pat.testumg.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.kappich.pat.testumg.util.connections;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Arrays;
import java.util.concurrent.ArrayBlockingQueue;

/**
 * @author Kappich Systemberatung
 * @version $Revision$
 */
public class DelayOutputStream extends OutputStream {
	private final OutputStream _inner;
	private final ArrayBlockingQueue<ByteArray> _queue = new ArrayBlockingQueue<ByteArray>(128);
	private int _maxPacketSize;
	private long _minPacketDelay;
	private long _transmissionDelay;
	public volatile static boolean DISABLE_DELAY = false;

	public DelayOutputStream(OutputStream outputStream, final long transmissionDelay, final double maxFlowRate) {
		_inner = outputStream;
		_transmissionDelay = transmissionDelay;
		if(maxFlowRate >= 1000000000){
			_minPacketDelay = 0;
			_maxPacketSize = Integer.MAX_VALUE;
		}
		else {
			_minPacketDelay = Math.max(transmissionDelay, 4);
			double doubleSize = maxFlowRate * _minPacketDelay / 1000.0;
			while(doubleSize < 128){
				_minPacketDelay *= 2;
				doubleSize = maxFlowRate * _minPacketDelay / 1000.0;
			}
			_maxPacketSize = (int) doubleSize;
		}
		new Thread(){
			@Override
			public void run() {
				try {
					while(true) {
						ByteArray byteArray = _queue.take();
						if(!DISABLE_DELAY) {
							long d = byteArray.getExpires() - System.currentTimeMillis();
							while(d > 0) {
								Thread.sleep(d);
								d = byteArray.getExpires() - System.currentTimeMillis();
							}
						}
						_inner.write(byteArray.getBytes());
						if(!DISABLE_DELAY) {
							Thread.sleep(_minPacketDelay);
						}
					}
				}
				catch(Exception e) {
					return;
				}
			}
		}.start();
	}

	@Override
	public void write(final int b) throws IOException {
		byte[] bytes = {(byte) b};
		enqueue(bytes);
	}

	@Override
	public void write(final byte[] b) throws IOException {
		enqueue(b.clone());
	}

	@Override
	public void write(final byte[] b, final int off, final int len) throws IOException {
		enqueue(Arrays.copyOfRange(b, off, off + len));
	}

	protected void enqueue(final byte[] bytes) throws IOException {
		if(bytes.length < _maxPacketSize){
			try {
				_queue.put(new ByteArray(bytes));
			}
			catch(InterruptedException e) {
				throw new IOException(e);
			}
		}
		else {
			for(int s = 0; s < bytes.length; s+=_maxPacketSize){
				try {
					_queue.put(new ByteArray(Arrays.copyOfRange(bytes, s, Math.min(s + _maxPacketSize, bytes.length))));
				}
				catch(InterruptedException e) {
					throw new IOException(e);
				}
			}
		}
	}

	private class ByteArray  {
		private final byte[] _bytes;
		private final long _expires;

		public ByteArray(final byte[] bytes) {
			_bytes = bytes;
			_expires = System.currentTimeMillis() + _transmissionDelay;
		}

		private byte[] getBytes() {
			return _bytes;
		}

		private long getExpires() {
			return _expires;
		}
	}
}
