/*
 * Copyright 2011 by Kappich Systemberatung Aachen
 *
 * This file is part of de.kappich.pat.testumg.
 *
 * de.kappich.pat.testumg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.kappich.pat.testumg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.kappich.pat.testumg.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.kappich.pat.testumg.util.connections;

import de.bsvrz.dav.daf.communication.lowLevel.ConnectionInterface;
import de.bsvrz.dav.daf.communication.lowLevel.telegrams.ApplicationDataTelegram;
import de.bsvrz.dav.daf.communication.lowLevel.telegrams.DataTelegram;
import de.bsvrz.dav.daf.communication.lowLevel.telegrams.KeepAliveTelegram;
import de.bsvrz.dav.daf.communication.tcpCommunication.TCP_IP_Communication;
import de.bsvrz.dav.daf.main.ConnectionException;

import java.io.InputStream;
import java.io.OutputStream;
import java.net.Socket;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArraySet;

/**
 * @author Kappich Systemberatung
 * @version $Revision$
 */
public class LocalDavDavConnection implements ConnectionInterface {

	private TCP_IP_Communication _communication;

	private int _subAdressNumber = 0;

	private final MultiDavProtocolParameter _parameters;

	private static final Set<String> DISABLED_CONNECTIONS = new CopyOnWriteArraySet<String>();

	private static final Map<String, ConnectionDelayInformation> CONNECTION_DELAYS = new ConcurrentHashMap<String, ConnectionDelayInformation>();

	static final Set<LocalDavDavConnection> _localDavDavConnections = new CopyOnWriteArraySet<LocalDavDavConnection>();

	public LocalDavDavConnection(final Socket socket, final String parameters) {
		_parameters = new MultiDavProtocolParameter(parameters);
		_communication = new TCP_IP_Communication(socket);
		_localDavDavConnections.add(this);
	}

	public LocalDavDavConnection(final String parameters) {
		_parameters = new MultiDavProtocolParameter(parameters);
		_communication = new TCP_IP_Communication();
		_localDavDavConnections.add(this);
	}

	public LocalDavDavConnection() {
		_parameters = new MultiDavProtocolParameter("");
		_communication = new TCP_IP_Communication();
	}

	public String getMainAdress() {
		return _communication.getMainAdress();
	}

	public int getSubAdressNumber() {
		return _communication.getSubAdressNumber();
	}

	public OutputStream getOutputStream() {
		ConnectionDelayInformation delayInformation = CONNECTION_DELAYS.get(_parameters.getMyName() + ":" + _communication.getSubAdressNumber());
		if(delayInformation == null){
			return _communication.getOutputStream();
		}
		return new DelayOutputStream(_communication.getOutputStream(), delayInformation.getTransmissionDelay(), delayInformation.getMaxFlowRate());
	}

	@Override
	public String toString() {
		return _parameters.getMyName() + ">" + _parameters.getOtherName(_communication.getMainAdress(), _communication.getSubAdressNumber());
	}

	public InputStream getInputStream() {
		return _communication.getInputStream();
	}

	public void connect(final String mainAddress, final int subAddressNumber) throws ConnectionException {
		_subAdressNumber = subAddressNumber;

		if(DISABLED_CONNECTIONS.contains(makeAdressKey(_parameters.getMyName(), subAddressNumber))) {
			throw new ConnectionException("Verbindung derzeit deaktiviert");
		}
		else {
			_communication.connect(mainAddress, subAddressNumber);
		}
	}

	public void disconnect() {
		_communication.disconnect();
	}

	public boolean isConnected() {
		return _communication.isConnected();
	}

	@Override
	public boolean isLoopback() {
		return _communication.isLoopback();
	}

	private void refreshEnabled(){
		final boolean disabled = DISABLED_CONNECTIONS.contains(makeAdressKey(_parameters.getMyName(), _subAdressNumber));
		if(disabled && _communication.isConnected()){
			_communication.disconnect();
		}
	}

	public static void disableConnection(final String davName, final int targetPort){
		DISABLED_CONNECTIONS.add(makeAdressKey(davName, targetPort));
		for(final LocalDavDavConnection connection : _localDavDavConnections) {
			connection.refreshEnabled();
		}
	}

	public static void enableConnection(final String davName, final int targetPort){
		DISABLED_CONNECTIONS.remove(makeAdressKey(davName, targetPort));
		for(final LocalDavDavConnection connection : _localDavDavConnections) {
			connection.refreshEnabled();
		}
	}

	/**
	 * @param maxFlowRate Maximale Datenrate in Bytes/Sekunde (ungefähr)
	 * @param transmissionDelay Zeit in Millisekunden, die Daten für die Strecke vom Sender zum Empfänger benötigen
	 */
	public static void setConnectionDelay(final String fromDav, final int toPort, final long transmissionDelay, final double maxFlowRate){
		CONNECTION_DELAYS.put(fromDav + ":" + toPort, new ConnectionDelayInformation(maxFlowRate, transmissionDelay));
	}

	private static String makeAdressKey(final String davName, final int targetPort) {
		return davName + ":" + targetPort;
	}

	public static void enableAll() {
		DISABLED_CONNECTIONS.clear();
		for(final LocalDavDavConnection connection : _localDavDavConnections) {
			connection.refreshEnabled();
		}
		_localDavDavConnections.clear();
	}

	/**
	 * Informationen über die geschwindigkeit einer Verbindung. Kann manuell gesetzt werden um Verbindungen testweise auszubremsen, etc.
	 */
	private static class ConnectionDelayInformation {
		private final long _transmissionDelay;
		private final double _maxFlowRate;

		/**
		 * @param maxFlowRate Maximale Datenrate in Bytes/Sekunde (ungefähr)
		 * @param transmissionDelay Zeit in Millisekunden, die Daten für die Strecke vom Sender zum Empfänger benötigen
		 */
		private ConnectionDelayInformation(final double maxFlowRate, final long transmissionDelay) {
			_maxFlowRate = maxFlowRate;
			_transmissionDelay = transmissionDelay;
		}

		private long getTransmissionDelay() {
			return _transmissionDelay;
		}

		private double getMaxFlowRate() {
			return _maxFlowRate;
		}
	}
}
